import Acl from 'acljs'
import { actionExists, getPermissions, Permissions } from './PermissionManager'
import { createImplement, invokeInterface } from 'sunny-js/util/module'
import { splitPath } from 'sunny-js/util/path'
import { RouteProps } from 'react-router'

export interface OperationProps {
  // 当前路由所属资源（resource）
  resource?: string
  // 当前路由组件对应的特权（privilege）
  privilege?: string
}

// 定义模块接口
interface ModuleInterfaces {
  getRole?: () => string
  defaultPrivilege?: string
  getOperation?: (route: RouteProps) => Required<OperationProps>
}
const iACL = {
  defaultPrivilege: 'view'
} as ModuleInterfaces
export const implement = createImplement(iACL)

export interface RouteLike {
  path: string,
  [key: string]: any
}
export type ACLRouteProps  = OperationProps & RouteLike

interface ACL {
  isAllowed: (role: string, resource: string, privilege: string) => boolean | never
}
let _acl: null | ACL = null
/**
 * 单例模式创建ACL实例，并初始化好权限
 * 该方法可以重复调用
 */
export function createACLInstance (permissions: Permissions): ACL {
  return _acl = _acl || new Acl(permissions)
}

export const getRole = invokeInterface(
  () => iACL.getRole,
  () => 'guest'
)
export const getOperation = invokeInterface(
  () => iACL.getOperation,
  (route: ACLRouteProps) => {
    // 先将path分割成片段
    const path = route.path
    const segments = splitPath(path)

    // 非后台管理类的path默认privilege为view（查看）
    let privilege: string = actionExists(path) ? segments.pop() : iACL.defaultPrivilege
    // 优先使用路由配置中的相关字段
    privilege = route.privilege || privilege

    let resource: string = segments.pop()
    resource = route.resource || resource

    return { resource, privilege }
  }
)

/**
 * 权限验证方法同步版本
 * Note：
 * 调用同步方法特别需要注意，ACL模块权限配置数据是否已经初始化完成
 */
export function isAllowedSync (route: ACLRouteProps) {
  if (!_acl) {
    throw new Error('You must initialize ACL Class first. Try to call ' + createACLInstance.name)
  }
  let allowed = false
  const role = getRole()
  const { resource, privilege } = getOperation(route)
  try {
    allowed = _acl.isAllowed(role, resource, privilege)
  } catch (e) {
    console.error(e)
  }
  console.debug(
    role,
    allowed ? 'is' : 'is not',
    'allowed to',
    (privilege||'?') + '(privilege)', (resource||'?') + '(resource)')
  return allowed
}

/**
 * 权限验证
 * 该方法时异步的
 * 1. 会先异步读取权限配置，然后缓存
 * 2. 再创建单例ACL Class实例
 * 3. 最后验证权限
 */
export function isAllowed (route: ACLRouteProps) {
  return getPermissions()
    .then(permissions => createACLInstance(permissions))
    .then(() => isAllowedSync(route) ? Promise.resolve() : Promise.reject())
    .catch(e => {
      console.error(e)
      return Promise.reject()
    })
}
